//****************************************************************************
//****************************************************************************
// OFORTH : MAIN
//****************************************************************************
//****************************************************************************

/*
  Copyright (c) 2014, 2018 Franck Bensusan. All rights reserved.

  This file is part of Oforth sources engine.

  Oforth is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation, either version 3
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, see http://www.gnu.org/licenses/.
*/

// ***************************************************************************

#include "oforth.h"

// Words compiler
extern void W_BYTECODE();
extern void W_CELLCODE();
extern void W_RAWCODE();
extern void W_RET();

// Words Primitivs
extern void W_checkStack();
extern void W_resolveContinue();
extern void W_resolveBreak();
extern void W_dup();
extern void W_drop();
extern void W_swap();
extern void W_over();
extern void W_rot();
extern void W_eq();
extern void W_pick();
extern void W_depth();
extern void W_class();

extern void W_col();
extern void W_implement();
extern void W_colend();
extern void W_to();
extern void W_doubleQuote();
extern void W_char();
extern void W_comment();
extern void W_immediate();
extern void W_isImmediate();
extern void W_forget();
extern void W_parsename();
extern void W_parsechar();
extern void W_nextchar();
extern void W_lastWord();
extern void W_strAsInteger();
extern void W_lit();
extern void W_litraw();
extern void W_closureVarUsed();
extern void W_closureVarSet();
extern void W_propAtt();
extern void W_setAtt();
extern void W_setImmAtt();
extern void W_setLongJump();

extern void W_mem();
extern void W_bye();
extern void W_errno();
extern void W_sleep();
extern void W_yield();
extern void W_sched();
extern void W_SystemTick();
extern void W_SystemLocalTime();
extern void W_interp();
extern void W_interpreter();

extern void W_st();
extern void W_t();

// Object 
extern void CM_Object_new();
extern void CM_Object_alloc();
extern void  M_Object_initialize();
extern void  M_Object_free();
extern void  M_Object_freeze();
extern void  M_Object_hashValue();

// Integers
extern fc integerInit;
extern void M_Integer_eq();
extern void M_Integer_cmp();
extern void M_Integer_add();
extern void M_Integer_sub();
extern void M_Integer_mult();
extern void M_Integer_div();
extern void M_Integer_mod();
extern void M_Integer_divmod();
extern void M_Integer_bitAnd();
extern void M_Integer_bitOr();
extern void M_Integer_bitXor();
extern void M_Integer_bitLeft();
extern void M_Integer_bitRight();
extern void M_Integer_asFloat();
extern void M_Integer_asDigit();

// Pointers 
extern void M_Integer_byteAt();
extern void M_Integer_bytePut();
extern void M_Integer_wordAt();
extern void M_Integer_wordPut();
extern void M_Integer_at();
extern void M_Integer_put();
extern void M_Integer_asString();

//Floats
extern void CM_Float_rand();
extern void M_Float_eq();
extern void M_Float_cmp();
extern void M_Float_add();
extern void M_Float_sub();
extern void M_Float_mult();
extern void M_Float_div();
extern void M_Float_sqrt();
extern void M_Float_powf();
extern void M_Float_ln();
extern void M_Float_cos();
extern void M_Float_sin();
extern void M_Float_tan();
extern void M_Float_atan();
extern void M_Float_asin();
extern void M_Float_acos();
extern void M_Float_asInteger();

// Object model
extern void  M_Symbol_size();

extern void CM_Word_find();
extern void  M_Word_name();
extern void  M_Word_forget();

extern void CM_Constant_new();
extern void  M_Constant_value();

extern void CM_Variable_new();
extern void  M_Variable_at();
extern void  M_Variable_put();

extern void CM_Alias_new();

extern void CM_Class_new();
extern void M_Class_addAtt();
extern void M_Class_findAtt();
extern void M_Class_implement();
extern void M_Class_classImplement();
extern void M_Class_is();

extern void CM_Property_new();
extern void M_Property_addAtt();
extern void M_Property_requires();
extern void M_Property_findAtt();

extern void M_Function_perform();
extern void M_Function_compile();
extern void M_Function_see();

extern void M_Implement_compile();
extern void CM_Method_new();
extern void M_Method_perform();
extern void M_Method_compile();
extern void M_Method_compileTOS();
extern void M_Method_compileSuper();
extern void M_Method_see();
extern void M_Method_seecl();

extern void CM_OSLib_new();
extern void CM_OSProc_new();
extern void  M_OSProc_perform();

extern void CM_Package_new();
extern void M_Package_imported();

extern void CM_Block_new();
extern void M_Block_perform();
extern void M_Block_compile();

extern void M_Exception_throw();

extern fc   arrayInit;
extern void M_Array_size();
extern void M_Array_forEachNext();
extern void M_Array_at();
extern void M_Array_add();
extern void M_Array_put();
extern void M_Array_last();
extern void M_Array_pop();
extern void M_Array_removeAt();
extern void M_Array_empty();

extern fc     bufferInit;
extern ffinal bufferFinalize;
extern void M_Buffer_byteSize();
extern void M_Buffer_byteAt();
extern void M_Buffer_bytePut();
extern void M_Buffer_utf8At();
extern void M_Buffer_forEachNext();
extern void M_Buffer_hashValue();
extern void M_Buffer_eq();
extern void M_Buffer_cmp();
extern void M_Buffer_evaluate();
extern void M_Buffer_perform();

extern fc stringInit;
extern void M_String_asSymbol();
extern void M_String_asFloat();
extern void M_String_forEachNext();
extern void M_String_load();
extern void M_String_empty();
extern void M_String_add();
extern void M_String_removeAt();
extern void M_String_removeLast();
extern void M_String_addSymbol();
extern void M_String_addBufferFormat();
extern void M_String_addFloatFormat();
extern void M_String_addIntegerFormat();

extern ffinal fileFinalize;
extern void CM_File_stats();
extern void M_File_open();
extern void M_File_isOpen();
extern void M_File_atEnd();
extern void M_File_close();
extern void M_File_position();
extern void M_File_reposition();
extern void M_File_flush();
extern void M_File_getChar();
extern void M_File_readWith();
extern void M_File_readLineWith();
extern void M_File_addChar();
extern void M_File_addSymbol();
extern void M_File_addBufferFormat();
extern void M_File_addFloatFormat();
extern void M_File_addIntegerFormat();

extern fc sysResourceInit;
extern void M_Resource_isOpen();
extern void M_Resource_close();

extern ffinal channelFinalize;
extern void M_Channel_initialize();
extern void M_Channel_receiveTimeout();
extern void M_Channel_sendTimeout();

extern void M_Console_ekey();
extern void M_Console_receiveTimeout();

extern void M_Socket_initialize();
extern void M_Socket_close();
extern void M_SocketServer_initialize();
extern void M_SocketServer_acceptTimeout();
extern void M_SocketClient_connectMode();
extern void M_SocketClient_receiveTimeout();
extern void M_SocketClient_sendTimeout();
extern ffinal sysSocketFinalize;

extern void newWorker(CThread *);

extern MyImplement *findImplement(Method *, Class *);
extern MyImplement *findClassImplement(Method *, Class *);
extern void call_poly();

extern PCell allocDictCell(CThread *thread, vint n);
extern PCell allocDict(CThread *thread, vint n);
extern MyObject *allocDictObject(CThread *thread, Class *cl, vint n);
extern MyString *allocDictString(CThread *thread, char *s, vint allocate);
extern MyNatural *allocDictInteger(CThread *thread, vint size);
extern RawPage *allocRawMem(CThread *thread, vint size);
extern MyString *allocPersistentString(CThread *thread, vint size);

extern PTR allocMem(CThread *, vint, char **);
extern Raw *allocCells(CThread *thread, vint n, uchar **last);
extern void freeMem(CThread *, PTR);

extern void sysThreadExecute(CThread *thread);
extern void sysThreadException(CThread *thread);
extern void setPersistent(MyObject *);
extern void sysThread(CThread *thread);
extern MYLONG getTick(int mode, struct timeb *t);
extern void sysExceptionFromC(CThread *thread, char *message);
extern void sysExParamFromC(CThread *thread, char *message, void *);
extern void checkImmutable(register CThread *thread, MyObject *o);

//*****************************************************************************
// Global variables

// Memory & GC
CThread MAINTHREAD, *mainThread = &MAINTHREAD;
MyTask  MAINTASK,   *mainTask =   &MAINTASK;
TryNode MAINTRY,    *mainTry =    &MAINTRY;

vint maxMemory = HEAP_WARN;
vint gcTimer, gcMemory, gcTicksByStep, gcAppByStep, gcVerboseLevel;
TICK nextLocalGC, nextGlobalGC, nextInitTick;
vint GCLOCAL, GCGLOBAL;

vint genericVT = 0;
Cell *allNewVT;

// Methods
Method *throwMethod, *newMethod, *initMethod, *doesNotUnderstand; 
Method *logMeth, *performMeth, *compileMeth, *atMethod;

MyFunction *yieldFunc, *loadOforth; 


//*****************************************************************************
// VM parameters
int consoleMode = KO;
int interpretMode = KO, optimMod = OTRUE;
vint checkState = OFALSE;
vint nbProcs;
vint mainDsSize = DEFAULTDSMAIN;

//*****************************************************************************
// Code, Dictionary and Compiler variables
MyVariable *CURRENT, *ICODE, *_LIT, *_CONSOLE, *STATE, *_INTERPRET;
MyVariable *LASTCALL, *CPBEFORELASTCALL, *CPLASTCALL;
MyVariable *allBlocks;
MyVariable *SOURCE, *SOURCETYPE, *SOURCENAME, *SOURCELINE, *SOURCEINDEX;

MySymbol **allSymbols;
Class **allClasses;
vint nbClasses = 0;
MyPackage *allPackages, *oforthPackage;
MyWord *lastWord;

//*****************************************************************************
// Workers and Tasks 
OforthVM *ovm;
vint workerId = 1;
vint maxWorkers = 1;
CThread *lastThread;

//*****************************************************************************
// Input/Output variables
vint nbArguments = 0;
char argIndexes[MAXARGS];
uchar InputChar[0x80];
MyArray *allArgs, *allPath;

int tty = -1;
FILE *stdInput;
MyFile *stdoutFile, *stderrFile;

MyString *nextWord, *cmdStdin, *replStr = NULL;
char *inputFile = NULL;
char *cmd = NULL;

MyConsole *mainConsole = NULL;

//*****************************************************************************
// Other variables
MYFLOAT tickFrequency = 0.0;
MyNatural *nzero;
int tickInit;

#ifdef __DBG__
   int pwo = KO;
#endif

//*****************************************************************************
// Kernel classes
Class *Object, *Nil, *Integer, *Exception, *Buffer, *String, *Array; 
Class *Word, *Classe, *Implement, *Virtual; 
Class *Property, *Package, *Function, *Methode, *Alias, *Block;
Class *Symbol, *Constant, *Variable;
Class *File, *Channel, *Console;

#ifdef __FLOAT_ENABLE__
Class *Float;
#endif

#ifdef __TCP_ENABLE__
Class *TCPSocket, *TCPSocketClient, *TCPSocketServer;
#endif

#ifdef __DYN_ENABLE__
Class *OSLib, *OSProc;
#endif

//*****************************************************************************


//*****************************************************************************
vint hash_value(char *str)
// VALIDATED *************
{
register unsigned long hash;
int c;

   hash = tickInit;
   while ((c = *str++)) { hash = c + (hash << 6) + (hash << 16) - hash; }
   hash &= MAX_CINT;
   // dbg("Hash value = %d %x\n", hash, hash);
   return hash;
}
//*****************************************************************************
void reallocVars(CThread *thread)
//*******************************
{
MyTask *t;
vint oldsize, newsize, i;
PCell newvt, max;

   t = thread->currentTask;

   oldsize = (vint) t->vtvar[0];
   newvt = allocMem(thread, (oldsize + 1) * 2 * sizeof(Cell), (char **) &max);
   newsize = max - newvt - 1;
   newvt[0] = (Cell) newsize;

   for(i = 1; i <= newsize; i++) 
      if (i <= oldsize) newvt[i] = t->vtvar[i]; else newvt[i] = NULL;
      
   if (t->vtvar != (PCell) &genericVT) freeMem(thread, t->vtvar); 
   t->vtvar = newvt;

   // dbg("Reallocated vars : old = %d, new = %d\n", oldsize, newsize);
}
//*****************************************************************************
Cell getvar(CThread *thread, MyVariable *v)
//*****************************************
{
PCell vars;
vint index;
MyTask *t;
   
   t = thread->currentTask; 
   vars = t->vtvar; index = v->index;
   while ((vint) vars[0] < index ) { reallocVars(thread); vars = t->vtvar; }

   return vars[index];
}
//*****************************************************************************
void setvar(CThread *thread, MyVariable *v, Cell value)
//*****************************************************
{
PCell vars;
vint index;
MyTask *t;
   
   t = thread->currentTask;
   vars = t->vtvar; index = v->index;
   while ((vint) vars[0] < index ) { reallocVars(thread); vars = t->vtvar; }

   vars[index] = value;
}
//****************************************************************************
MySymbol *find_symbol(char *name, vint hash)
//******************************************
{
MySymbol *s;

   s = allSymbols[hash];
   while(s != NULL) {
      if (strcmp(name, symbuf(s)) == OK) return s;
      s = s->next;
      }
   return NULL;
}
//*****************************************************************************
MySymbol *add_symbol(CThread *thread, char *name)
// VALIDATED ************************************
{
MySymbol *s;
vint len, hash;
char c, d;

   len = strlen(name);
   if (len == 0 || len >= MAXNAMEWORD)
      sysExceptionFromC(thread, "Bad symbol name");

   c = *name; d = *(name+1);
   if (c >= '0' && c <= '9' && d >= '0' && d <= '9')
      sysExceptionFromC(thread, "Bad symbol name");

   hash = hash_value(name) % SYM_HASH_VALUE;
   s = (MySymbol *) find_symbol(name, hash); 

   if (s == NULL) {
      s = (MySymbol *) allocRawMem(thread, sizeof(MySymbol) + len + 1);
      if (Symbol != NULL) s->objptr = (Symbol->classPtr | MSKPERSISTENT) & MSKMODERUN;
      else s->objptr = 0;

      memcpy(symbuf(s), name, len + 1); 
      s->next = allSymbols[hash]; 
      allSymbols[hash] = s;
      }

   // dbg("Adding symbol %s\n", name);
   return s;
}
//*****************************************************************************
Class *objectClass(MyObject *o)
// VALIDATED ******************
{
   if(ISNIL(o)) return Nil;
   if(ISINT(o)) return Integer;
   return allClasses[o->objptr & MSKCLASSID]; 
}
//*****************************************************************************
vint isKindOf(MyObject *o, Class *parent)
// VALIDATED ****************************
{
Class *cl = objectClass(o);

   while(cl != NULL) {
      if (cl == parent) return OK;
      cl = cl->super;
      }

   return KO; 
}
//*****************************************************************************
MyWord *find_package_word(MyPackage *pack, MySymbol *name)
//********************************************************
{
vint hash;
MyWord *h;

   hash = hash_value(symbuf(name)) % PACK_HASH_VALUE;
   h = (MyWord *) pack->wrdlist->buf[hash];
   // dbg("name = %s, hash = %d h = %x\n", symbuf(name), hash, h);

   while(h != NULL && h->name != name) h = h->next; 

   if (h != NULL && objectClass((MyObject *) h) == Alias) 
      h = ((MyAlias *) h)->parent; 

   return h;
}
//*****************************************************************************
MyWord *wordFind(CThread *thread, uchar *name)
//**********************************************
{
MyPackage *pack;
MyImportedPack *import;
char *checkName, temp[MAXNAMEWORD];
vint i, hash;
MySymbol *s;
MyWord *h;

   checkName = temp;
   i = 0;
   pack = NULL;

   // dbg("Finding word : <%s>\n", name);
   while(*name != 0 && (*name != ':' || checkName == temp) && i < MAXNAMEWORD) 
      { *checkName++ = *name++; i ++; }
   if (i == MAXNAMEWORD) return NULL;

   // Checks if the name is qualified 
   if (*name == ':') {
      if (*(name + 1) != '\0') {
         *checkName = '\0';
         hash = hash_value(temp) % SYM_HASH_VALUE;
         if ((s = find_symbol(temp, hash)) == NULL) return NULL;

         pack = (MyPackage *) find_package_word(oforthPackage, s);
         if (pack == NULL || objectClass((MyObject *) pack) != Package) 
            return NULL;
         checkName = temp;
         name++; i = 0;
         while(*name != 0 && i < MAXNAMEWORD) { *checkName++ = *name++; i++; }
         if (i == MAXNAMEWORD) return NULL;
         }
      else { *checkName = ':'; checkName++; }
      }

   *checkName = '\0';

   hash = hash_value(temp) % SYM_HASH_VALUE;
   // dbg("temp = %s, hash = %d\n", temp, hash);
   if ((s = find_symbol(temp, hash)) == NULL) return NULL;

   // If qualified, returns object into package
   if (pack != NULL) return find_package_word(pack, s);

   h = find_package_word((MyPackage *) getvar(thread, CURRENT), s);
   if (h != NULL) return h;

   // dbg("Finding not qualified word : <%s> %x \n", temp, getvar(thread, CURRENT));
   import = ((MyPackage *) getvar(thread, CURRENT))->importedPackages;
   while(import != NULL) {
      h = find_package_word(import->package, s);
      if (h != NULL) return h;
      import = import->next;
      }
   
   if ((MyPackage *) getvar(thread, CURRENT) == oforthPackage) return NULL;

   // Else, try to find it into oforth package and imported packages
   h = find_package_word(oforthPackage, s); 
   if (h != NULL) return h;

   import = oforthPackage->importedPackages;
   while(import != NULL) {
      h = find_package_word(import->package, s);
      if (h != NULL) return h;
      import = import->next;
      }

   return NULL;
}
//*****************************************************************************
void addWordToCurrent(CThread *thread, MyWord *w)
//***********************************************
{
vint hash;
MyPackage *p;
Cell *buf;

   hash = hash_value(symbuf(w->name)) % PACK_HASH_VALUE;
   p = (MyPackage *) getvar(thread, CURRENT);
   buf = p->wrdlist->buf;
   w->next = (MyWord *) buf[hash];
   buf[hash] = (Cell) w;
}
//*****************************************************************************
MyWord *allocWord(CThread *thread, Class *cl, vint size, char *name)
// VALIDATED *******************************************************
{
MySymbol *s;
MyWord *h;

   // dbg("Adding symbol : %s\n", name);
   s = add_symbol(thread, name);
   
   if (find_package_word((MyPackage *) getvar(thread, CURRENT), s) != NULL) 
      sysExParamFromC(thread, "Can't redefine word", s);

   h = (MyWord *) allocDictObject(thread, cl, size);
   h->name = s;
   addWordToCurrent(thread, h);
   lastWord = h;
   return h;
}
//*****************************************************************************
Method *method(CThread *thread, char *name)
// VALIDATED ******************************
{
Method *m;

   if ((m = (Method *) wordFind(thread, (uchar *) name)) != NULL) return m;
   m = (Method *) allocWord(thread, Methode, sizeof(Method), name);
   m->vt = (PCell) &genericVT; 
   return m;
}
//******************************************************************************
MyImplement *add_impl(CThread *thread, Method *m, Class *inst, Class *cl, fptr f, Class *type)
// VALIDATED ****************************************************************
{
MyImplement **pn, *impl, *pi = NULL;
vint j;
PCell vt = NULL;

   if (inst == Classe) {
      pn = &(m->classeList); pi = findClassImplement(m, cl); 
      if (m == newMethod) vt = allNewVT; 
      } 
   else { 
      pn = &(m->objectList); vt = m->vt; 
      if (objectClass((MyObject *) cl) == Classe) pi = findImplement(m, cl);
      }

   while (*pn != NULL) {
      if ((*pn)->implement == cl)
         sysExParamFromC(thread,"Method already defined for",cl);
      pn = &((*pn)->next);
      }

   if (pi != NULL && objectClass((MyObject *) pi) == Implement) 
      sysExParamFromC(thread, "Can't redefine non virtual method", m);

   if (vt != NULL) { for(j = 1; j <= (vint) *vt; j++) vt[j] = (Cell) call_poly; }

   *pn = impl = (MyImplement *) allocDictObject(thread, type, sizeof(MyImplement));
   impl->next = NULL;
   impl->code = f;
   impl->implement = cl; 
   return impl;
}
//*****************************************************************************
void add_method(Class *inst, Class *cl, char *name, fptr f, Class *type)
//**********************************************************************
{
   add_impl(mainThread, method(mainThread, name), inst, cl, f, type);
}
//*****************************************************************************
void implement(Class *cl, char *name, fptr f)
//*******************************************
{
   add_method(Object, cl, name, f, Implement);
}
//*****************************************************************************
void classImplement(Class *cl, char *name, fptr f)
//************************************************
{
   add_method(Classe, cl, name, f, Implement);
}
//*****************************************************************************
void virtual(Class *cl, char *name, fptr f)
//*****************************************
{
   add_method(Object, cl, name, f, Virtual);
}
//*****************************************************************************
MyConstant *constant(CThread *thread, char *name, Cell value)
// VALIDATED ************************************************
{
MyConstant *c;

   // dbg("New constant <%s>, value <%x>\n", name, value);
   c = (MyConstant *) allocWord(thread, Constant, sizeof(MyConstant), name);
   checkImmutable(thread, (MyObject *) value);
   setPersistent((MyObject *) value);
   c->value = value;

   // dbg("New constant allocated <%s>, value <%x>\n", name, value);
   return c;
}
//*****************************************************************************
MyVariable *variable(CThread *thread, char *name)
// VALIDATED ************************************
{
MyVariable *v;

   // dbg("Adding variable : %s\n", name);
   v = (MyVariable *) allocWord(thread, Variable, sizeof(MyVariable), name);
   v->index = LockInc(thread->vm->nbVars);
   // dbg("New variable %s at index %d\n", name, v->index);
   return v;
}
//********************************************************************************
MyFunction *primitive(char *name, void *fct, vint type)
// VALIDATED ******************************************
{
MyFunction *f;

   f = (MyFunction *) allocWord(mainThread, Function, sizeof(MyFunction), name);
   f->objptr |= type;
   f->code = fct;
   return f;
}
//********************************************************************************
Class *class(CThread *thread, char *name, Class *super)
//*****************************************************
{
vint i;
Class *cl, **old;

   if (nbClasses == MAX_CLASSES) 
      { sysExceptionFromC(thread, "Max number of classes reached"); }

   if (nbClasses == MIN_CLASSES - 1) {
      old = allClasses;
      allClasses = (Class **) allocRawMem(thread, MAX_CLASSES * sizeof(Cell));
      memcpy(allClasses, old, MIN_CLASSES * sizeof(Cell));

      allNewVT   = (Cell *) allocRawMem(thread, MAX_CLASSES * sizeof(Cell));
      allNewVT[0] = (Cell) (MAX_CLASSES - 1);
      for(i = 1; i < MAX_CLASSES; i++) allNewVT[i] = (Cell) call_poly;
      }
   
   cl = (Class *) allocWord(thread, Classe, sizeof(Class), name); 
   nbClasses++; 
   allClasses[nbClasses] = cl;
   cl->super = super; 
   
   if (super == NULL) {
      cl->size = 1; 
      cl->classPtr = 1 + MSKMODEINIT; 
      }
   else {
      cl->size          = super->size;
      cl->initFunction  = super->initFunction;
      cl->finalFunction = super->finalFunction;
      cl->classPtr      = nbClasses + (super->classPtr & MSKSTATE);
      cl->classPtr      &= ~MSKNOTATTGC;
      }

	return cl;
}
//******************************************************************************
Class *fclass(CThread *thread, char *name, Class *super, vint hid, vint flags)
//****************************************************************************
{
Class *cl;

   cl = class(thread, name, super);
   cl->size += hid;
   cl->classPtr += hid * (MSKATT + MSKATTHIDDEN);
   cl->classPtr |= flags;
   return cl;
}
//*****************************************************************************
Attribute *add_attribute(CThread *thread, Class *cl, MySymbol *name, MyProperty *p, vint mutable)
// VALIDATED ************************************************************
{
Attribute *na;
   
   if (cl->size >= MAXATTRIBUTE) 
      { sysExceptionFromC(thread, "Max number of attributes reached"); }

   na = (Attribute *) allocRawMem(thread, sizeof(Attribute));
   na->name = name;
   na->property = p;
   na->isMutable = mutable;
   na->next = cl->attList; 

   cl->attList = na;
   cl->size++;
   cl->classPtr += MSKATT; 
   if (mutable == OTRUE) cl->classPtr |= MSKMUTABLE;

   return na;
}
//*****************************************************************************
void attribute(Class *cl, char *name)
// VALIDATED ***********************
{
   add_attribute(mainThread, cl, add_symbol(mainThread, name), NULL, OFALSE);
}
//*****************************************************************************
void mutable(Class *cl, char *name)
// VALIDATED **********************
{
   add_attribute(mainThread, cl, add_symbol(mainThread, name), NULL, OTRUE);
}
//******************************************************************************
void init_image(CThread *thread, int h, int show, int argc, char **argv)
//**********************************************************************
{
Class *cl, *obj, *coll, *ress;
vint i;
Attribute *na;
MyArray *w;

   allSymbols  = (MySymbol **) allocRawMem(thread, SYM_HASH_VALUE * sizeof(Cell));

   allClasses  = (Class **)    allocRawMem(thread, MIN_CLASSES * sizeof(Cell));
   allNewVT    = (Cell *)      allocRawMem(thread, MIN_CLASSES * sizeof(Cell));
   allNewVT[0] = (Cell) (MIN_CLASSES - 1);
   for(i = 1; i < MIN_CLASSES; i++) allNewVT[i] = (Cell) call_poly;
   
   // CURRENT variable and oforth package : first word
   CURRENT = (MyVariable *) allocDictObject(thread, NULL, sizeof(MyVariable));
   CURRENT->index = 1;

   ovm->nbVars = 2;
   mainTask->vtvar = (PCell) allocMem(thread, 2 * sizeof(Cell), NULL);
   mainTask->vtvar[0] = (Cell) 1;

   oforthPackage = (MyPackage *) allocDictObject(thread, NULL, sizeof(MyPackage));
   oforthPackage->name = add_symbol(thread, "oforth");
   oforthPackage->next = NULL;
   oforthPackage->wrdlist = w = (MyArray *) allocDictObject(thread,NULL,sizeof(MyArray));
   w->buf = (Cell *) allocDictCell(thread, PACK_HASH_VALUE);
   w->free = w->alloc = w->buf + PACK_HASH_VALUE;

   setvar(thread, CURRENT, (Cell) oforthPackage);
   addWordToCurrent(thread, (MyWord *) oforthPackage);
   
   // Built-in classes

   Object = obj = class(thread, "Object", NULL);
   Nil = class(thread, "Null", obj); 
#ifdef __DBG__   
   if (nbClasses != 2) dbg("NUll = %d, id = 2\n", nbClasses); 
#endif

   Integer = fclass(thread, "Integer", obj, 3, MSKFINALIZE + MSKNOTATTGC);
   Integer->initFunction = integerInit;
   Integer->finalFunction = bufferFinalize;
#ifdef __DBG__   
   if (nbClasses != 3) dbg("Integer = %d, id = 3\n", nbClasses); 
#endif

   Word = fclass(thread, "Word", obj, 2, 0);
   
   Classe = fclass(thread, "Class", Word, 6, 0);
   attribute(Classe, "superclass");

#ifdef __DBG__   
   if (nbClasses != 5) dbg("Class = %d, id = 5\n", nbClasses); 
#endif

   for(i = 1; i <= nbClasses; i++)
      ((Class *) allClasses[i])->objptr 
         = (Classe->classPtr | MSKPERSISTENT) & MSKMODERUN;
         
#ifdef __FLOAT_ENABLE__
   Float = fclass(thread, "Float", obj, FLOATSIZE - 1, MSKNOTATTGC);
#ifdef __DBG__   
   if (nbClasses != 6) dbg("Float = %d, id = 6\n", nbClasses); 
#endif
#endif

   Property = fclass(thread, "Property", Word, 2, 0);

   Package = fclass(thread, "Package", Word, 2, 0);
   oforthPackage->objptr = (Package->classPtr | MSKPERSISTENT) & MSKMODERUN;

   Function = fclass(thread, "Function", Word, 1, 0);

   Methode = fclass(thread, "Method", Word, 1, 0);
   attribute(Methode, "implements");
   attribute(Methode, "classImplements");

   Constant = class(thread, "Constant", Word);
   attribute(Constant, "value");

   Alias =      class(thread, "Alias", Word);
   Variable  = fclass(thread, "TVar",  Word, 1, 0);

   Symbol = fclass(thread, "Symbol", obj, 2, MSKPERSISTENT);
   oforthPackage->name->objptr = Symbol->classPtr;

   for(i = 1; i <= nbClasses; i++) {
      cl = allClasses[i];
      cl->name->objptr = Symbol->classPtr;

      for(na = cl->attList; na != NULL; na = na->next)
         na->name->objptr = Symbol->classPtr;
      }

   Block   = fclass(thread, "Block",   obj, 3, 0);

   Implement = fclass(thread, "Implement", obj, 2, 0);
   attribute(Implement, "implementor");

   Virtual = class(thread, "Virtual", Implement);

   coll = class(thread, "Collection", obj);

   Buffer = fclass(thread, "Buffer", coll, 3, 
                     MSKBUFFER + MSKFINALIZE + MSKNOTATTGC + MSKMUTABLE);
   Buffer->initFunction = bufferInit;
   Buffer->finalFunction = bufferFinalize;

   String = fclass(thread, "String", Buffer, 0, MSKNOTATTGC + MSKMUTABLE);
   String->initFunction = stringInit;

   Array = fclass(thread, "Array", coll, 3, 
                     MSKBUFFER + MSKFINALIZE + MSKNOTATTGC + MSKMUTABLE);
   Array->initFunction = arrayInit;
   Array->finalFunction = bufferFinalize;

   oforthPackage->wrdlist->objptr = (Array->classPtr | MSKPERSISTENT) & MSKMODERUN;

   Exception   = class(thread, "Exception", obj);
   attribute(Exception, "message");
   mutable(Exception, "object");

   File = fclass(thread, "File", obj, 1, MSKBUFFER + MSKFINALIZE); 
   attribute(File, "name");
   attribute(File, "mode");
   File->finalFunction = fileFinalize;

   ress = fclass(thread, "Resource", obj, 5, 0);
   ress->initFunction = sysResourceInit;

   Channel = fclass(thread, "Channel", ress, 7, MSKFINALIZE);
   Channel->finalFunction = channelFinalize;

   Console = class(thread, "Console", ress);
   mutable(Console, "cmd");
   mutable(Console, "xcursor");
   mutable(Console, "history");

#ifdef __DYN_ENABLE__
   OSLib       = fclass(thread, "DynLib", Word, 2, 0);
   attribute(OSLib, "fileName");

   OSProc      = fclass(thread, "DynProc", Word, 2, 0);
   attribute(OSProc, "lib");
   attribute(OSProc, "procName");
   attribute(OSProc, "return");
#endif

#ifdef __TCP_ENABLE__
   TCPSocket = fclass(thread, "TCPSocket", ress, 6, MSKFINALIZE);
   attribute(TCPSocket, "port");
   TCPSocket->finalFunction = sysSocketFinalize;

   TCPSocketServer = class(thread, "TCPSocketServer", TCPSocket);
   attribute(TCPSocketServer, "mode");
   attribute(TCPSocketServer, "backlog");

   TCPSocketClient = class(thread, "TCPSocketClient", TCPSocket);
   attribute(TCPSocketClient, "host");
   attribute(TCPSocketClient, "remotePort");
#endif
   
#ifdef __DBG__   
   // dbg("Number of built in classes : %d\n", nbClasses); 
#endif

   CURRENT->objptr = (Variable->classPtr | MSKPERSISTENT) & MSKMODERUN;
   CURRENT->name = add_symbol(thread, "CURRENT");
   addWordToCurrent(thread, (MyWord *) CURRENT);

   ICODE      = variable(thread, "CODE");
   _LIT       = variable(thread, "_LIT");
   _CONSOLE   = variable(thread, "System.Console");
   _INTERPRET = variable(thread, "_INTERPRET");

   // Compiler primitives
   primitive("_BYTECODE",        W_BYTECODE, 0); 
   primitive("_CELLCODE",        W_CELLCODE, 0); 
   primitive("_RAWCODE",         W_RAWCODE, 0); 
   primitive("_RET",             W_RET, 0); 
   primitive("_header",          W_col, 0);
   primitive("_;",               W_colend, MSKIMMEDIATE);
   primitive("_implement",       W_implement, 0);
   primitive("_resolveContinue", W_resolveContinue, 0);
   primitive("_resolveBreak",    W_resolveBreak, 0);
   primitive("_closureVarUsed",  W_closureVarUsed, 0);
   primitive("_closureVarSet",   W_closureVarSet, 0);
   primitive("_propAtt",         W_propAtt, 0);
   primitive("_setAtt",          W_setAtt, 0);
   primitive("_setImmAtt",       W_setImmAtt, 0);
   primitive("_setLongJump",     W_setLongJump, 0);
   primitive("_strAsInteger",    W_strAsInteger, 0);
   primitive("_lit",             W_lit, 0);
   primitive("_checkStack",      W_checkStack, 0); 

   // Primitives
   primitive("dup",              W_dup, 0); 
   primitive("drop",             W_drop, 0); 
   primitive("swap",             W_swap, 0); 
   primitive("over",             W_over, 0); 
   primitive("rot",              W_rot, 0); 
   primitive("pick",             W_pick, 0); 
   primitive("=",                W_eq, 0); 
   primitive(".depth",           W_depth, 0); 
   primitive("class",            W_class, 0); 

   primitive("to",               W_to, MSKIMMEDIATE);
   primitive("\"",               W_doubleQuote, MSKIMMEDIATE);
   primitive("\'",               W_char, MSKIMMEDIATE);
   primitive("\\",               W_comment, MSKIMMEDIATE); 
   primitive("parse-name",       W_parsename, 0); 
   primitive("parse-char",       W_parsechar, 0); 
   primitive("next-char",        W_nextchar, 0); 
   primitive("LAST-WORD",        W_lastWord, 0); 
   primitive("immediate",        W_immediate, 0);
   primitive("immediate?",       W_isImmediate, 0);
   primitive("forget:",          W_forget, 0);
   setvar(thread, _LIT, (Cell) primitive("_litraw", W_litraw, 0));
   setvar(thread, _INTERPRET, (Cell) primitive("_interp", W_interp, 0));

   primitive("bye",              W_bye, 0);
   primitive("mem",              W_mem, 0);
   primitive("sleep",            W_sleep, 0);
   primitive("sched",            W_sched, 0);
   primitive("System.tick",      W_SystemTick, 0);
   primitive("System.localTime", W_SystemLocalTime, 0);

   yieldFunc = primitive("yield", W_yield, 0);
   loadOforth = primitive("interpretTask",  W_interpreter, 0); 

#ifdef __DBG__
   primitive(".st",              W_st, 0); 
   primitive(".t",               W_t, 0); 
#endif

   // Methods
   doesNotUnderstand = method(thread, "doesNotUnderstand");
   throwMethod       = method(thread, "throw");
   logMeth           = method(thread, "log");
   newMethod         = method(thread, "new"); 
   atMethod          = method(thread, "at");
   initMethod        = method(thread, "initialize");
   performMeth       = method(thread, "perform");
   compileMeth       = method(thread, "_compile");

   // Implementations
   add_method(Classe, obj, "new",            CM_Object_new, Virtual);
   add_method(Classe, obj, "alloc",          CM_Object_alloc, Virtual);
          virtual(obj,     "initialize",      M_Object_initialize);
          virtual(obj,     "free",            M_Object_free);
          virtual(obj,     "hashValue",       M_Object_hashValue);
        implement(obj,     "freeze",          M_Object_freeze);

        implement(Symbol,  "size", M_Symbol_size);

   classImplement(Word,     "find",          CM_Word_find);
        implement(Word,     "name",           M_Word_name);
        implement(Word,     "forget",         M_Word_forget);

   classImplement(Constant, "new",           CM_Constant_new);
        implement(Constant, "value",          M_Constant_value);

   classImplement(Variable, "new",           CM_Variable_new);
        implement(Variable, "at",             M_Variable_at);
        implement(Variable, "put",            M_Variable_put);

   classImplement(Classe,   "new",           CM_Class_new); 
        implement(Classe,   "_findAtt",       M_Class_findAtt);
        implement(Classe,   "_addAtt",        M_Class_addAtt);
        implement(Classe,   "implement",      M_Class_implement);
        implement(Classe,   "classImplement", M_Class_classImplement);
        implement(Classe,   "is:",            M_Class_is);

   classImplement(Property, "new",           CM_Property_new);
        implement(Property, "_findAtt",       M_Property_findAtt);
        implement(Property, "_addAtt",        M_Property_addAtt);
        implement(Property, "requires:",      M_Property_requires);

   classImplement(Alias,    "new",           CM_Alias_new);

        implement(Function, "_compile",       M_Function_compile);
        implement(Function, "perform",        M_Function_perform);

   classImplement(Methode,  "new",           CM_Method_new);
        implement(Implement, "_compile",      M_Implement_compile);
        implement(Methode,  "perform",        M_Method_perform);
        implement(Methode,  "_compile",       M_Method_compile);
        implement(Methode,  "_compileTOS",    M_Method_compileTOS);
        implement(Methode,  "_compileSuper",  M_Method_compileSuper);

#ifdef __DBG__
        implement(Function, "see",            M_Function_see);
        implement(Methode,  "see",            M_Method_see);
        implement(Methode,  "seecl",          M_Method_seecl);
#endif

   classImplement(Package,  "new",           CM_Package_new);
        implement(Package,  "_imported",      M_Package_imported);

   // Blocks
   classImplement(Block,    "new",           CM_Block_new);
        implement(Block,    "compile",        M_Block_compile);
        implement(Block,    "perform",        M_Block_perform);

   // Integers 
        implement(Integer, "==",              M_Integer_eq);
        implement(Integer, "<=",              M_Integer_cmp);
        implement(Integer, "+",               M_Integer_add);
        implement(Integer, "-",               M_Integer_sub);
        implement(Integer, "*",               M_Integer_mult);
        implement(Integer, "/",               M_Integer_div);
        implement(Integer, "mod",             M_Integer_mod);
        implement(Integer, "/mod",            M_Integer_divmod);
        implement(Integer, "bitAnd",          M_Integer_bitAnd);
        implement(Integer, "bitOr",           M_Integer_bitOr);
        implement(Integer, "bitXor",          M_Integer_bitXor);
        implement(Integer, "bitLeft",         M_Integer_bitLeft);
        implement(Integer, "bitRight",        M_Integer_bitRight);
        implement(Integer, ">digit",          M_Integer_asDigit);

   // Ptr
        implement(Integer, "_>string",        M_Integer_asString);
        implement(Integer, "_byteAt",         M_Integer_byteAt);
        implement(Integer, "_bytePut",        M_Integer_bytePut);
        implement(Integer, "_wordAt",         M_Integer_wordAt);
        implement(Integer, "_wordPut",        M_Integer_wordPut);
        // implement(Integer, "_int32At",        M_Integer_at);
        // implement(Integer, "_int32Put",       M_Integer_put);

        implement(Integer, "_at",             M_Integer_at);
        implement(Integer, "_put",            M_Integer_put);

#ifdef __FLOAT_ENABLE__
        implement(Integer, ">float",          M_Integer_asFloat);

   classImplement(Float, "rand",             CM_Float_rand);
        implement(Float, "==",                M_Float_eq);
        implement(Float, "<=",                M_Float_cmp);
        implement(Float, "+",                 M_Float_add);
        implement(Float, "-",                 M_Float_sub);
        implement(Float, "*",                 M_Float_mult);
        implement(Float, "/",                 M_Float_div);
        implement(Float, "sqrt",              M_Float_sqrt);
        implement(Float, "powf",              M_Float_powf);
        implement(Float, "ln",                M_Float_ln);
        implement(Float, "cos",               M_Float_cos);
        implement(Float, "sin",               M_Float_sin);
        implement(Float, "tan",               M_Float_tan);
        implement(Float, "atan",              M_Float_atan);
        implement(Float, "acos",              M_Float_acos);
        implement(Float, "asin",              M_Float_asin);
        implement(Float, ">integer",          M_Float_asInteger);
#endif

   // Exceptions
        implement(Exception, "throw",         M_Exception_throw);

   // Array
        implement(Array, "size",              M_Array_size);
        implement(Array, "forEachNext",       M_Array_forEachNext);
        implement(Array, "at",                M_Array_at);
          virtual(Array, "add",               M_Array_add);
          virtual(Array, "put",               M_Array_put);
        implement(Array, "last",              M_Array_last);
        implement(Array, "pop",               M_Array_pop);
        implement(Array, "removeAt",          M_Array_removeAt);
        implement(Array, "empty",             M_Array_empty);

   // Buffer
        implement(Buffer, "byteSize",         M_Buffer_byteSize);
        implement(Buffer, "byteAt",           M_Buffer_byteAt);
        implement(Buffer, "bytePut",          M_Buffer_bytePut);
        implement(Buffer, "utf8At",           M_Buffer_utf8At);
          virtual(Buffer, "forEachNext",      M_Buffer_forEachNext);
        implement(Buffer, "evaluate",         M_Buffer_evaluate);
        implement(Buffer, "perform",          M_Buffer_perform);
        implement(Buffer, "==",               M_Buffer_eq);
        implement(Buffer, "<=",               M_Buffer_cmp);
        implement(Buffer, "hashValue",        M_Buffer_hashValue);

   // Strings
        implement(String, "removeAt",         M_String_removeAt);
        implement(String, "removeLast",       M_String_removeLast);
        implement(String, "empty",            M_String_empty);
        implement(String, "add",              M_String_add);
        implement(String, "addChar",          M_String_add);
        implement(String, "addSymbol",        M_String_addSymbol);
        implement(String, "addIntegerFormat", M_String_addIntegerFormat);
        implement(String, "addBufferFormat",  M_String_addBufferFormat);
        implement(String, "forEachNext",      M_String_forEachNext);
        implement(String, "load",             M_String_load);
        implement(String, ">symbol",          M_String_asSymbol);
        implement(String, ">float",           M_String_asFloat);
#ifdef __FLOAT_ENABLE__
        implement(String, "addFloatFormat",   M_String_addFloatFormat);
#endif

   // Files
   classImplement(File, "stats",             CM_File_stats);
        implement(File, "open",               M_File_open); 
        implement(File, "open?",              M_File_isOpen);
        implement(File, "end?",               M_File_atEnd);
        implement(File, "close",              M_File_close); 
        implement(File, "position",           M_File_position);
        implement(File, "reposition",         M_File_reposition);
        implement(File, "flush",              M_File_flush);
        implement(File, "readLineWith",       M_File_readLineWith);
        implement(File, "readWith",           M_File_readWith);
        implement(File, ">>",                 M_File_getChar);
        implement(File, "addChar",            M_File_addChar); 
        implement(File, "addSymbol",          M_File_addSymbol);
        implement(File, "addIntegerFormat",   M_File_addIntegerFormat);
        implement(File, "addBufferFormat",    M_File_addBufferFormat);
#ifdef __FLOAT_ENABLE__
        implement(File, "addFloatFormat",     M_File_addFloatFormat);
#endif


   // Resources
        implement(ress, "open?",              M_Resource_isOpen); 
          virtual(ress, "close",              M_Resource_close);

        implement(Channel, "initialize",      M_Channel_initialize); 
        implement(Channel, "receiveTimeout",  M_Channel_receiveTimeout);
        implement(Channel, "sendTimeout",     M_Channel_sendTimeout);

        implement(Console, "ekey?",           M_Console_ekey);
        implement(Console, "receiveTimeout",  M_Console_receiveTimeout);

#ifdef __DYN_ENABLE__
   classImplement(OSLib,    "new",           CM_OSLib_new);

   classImplement(OSProc,   "new",           CM_OSProc_new);
        implement(OSProc,   "perform",        M_OSProc_perform);
#endif

#ifdef __TCP_ENABLE__
          virtual(TCPSocket, "initialize",    M_Socket_initialize);
        implement(TCPSocket, "close",         M_Socket_close); 

        implement(TCPSocketServer, "initialize", M_SocketServer_initialize); 
        implement(TCPSocketServer, "_acceptTimeout", M_SocketServer_acceptTimeout);
   
        implement(TCPSocketClient, "connectMode", M_SocketClient_connectMode);
        implement(TCPSocketClient, "receiveWithTimeout",M_SocketClient_receiveTimeout);
        implement(TCPSocketClient, "sendTimeout", M_SocketClient_sendTimeout);
#endif

   // dbg("Number of built-in classes = %d\n", nbClasses); 

}
//*******************************************************************
vint optvalue(char *str, char *option)
// VALIDATED *************************
{
char *endptr;
long val;

   errno = 0;
   val = strtol(str, &endptr, 10);
   
   if (errno != 0 || endptr == str || val < 0) 
      { printf("Integer value required for --%s\n", option); exit(KO); }

   return (vint) val;
}
//*****************************************************************************
int ret_args(int argc, char **argv) 
// VALIDATED **********************
{
int ch, ch1 = 0;
int optind = 0, i;
char *optarg, *cp;

   for (i = 0; i < MAXARGS; i++) argIndexes[i] = '\0';

   // dbg("otpind = %d, argc = %d\n", optind, argc);
   while(optind < argc) {
      cp = argv[optind];
      // dbg("cp = %s\n", cp);
      if (cp[0] == '-' && cp[1] == '-') {
        ch = cp[2]; optarg = cp + 3;
        if (ch == 'X') { ch1 = cp[3]; optarg = cp + 4; }
        switch(ch) {
           case 'i' : interpretMode = OK; break;
           case 't' : checkState = OTRUE; break;
           case 'S' : mainDsSize = optvalue(optarg, "S"); break;
           case 'P' : cmd = optarg; break;
           case 'M' : maxMemory  = optvalue(optarg, "M"); break;

#ifdef __MTHREAD_ENABLE__
           case 'W' : maxWorkers = optvalue(optarg, "W"); break;
#endif

#ifdef __DBG__
           case 'p' : pwo = OK; break;
           case 'C' : optimMod = OFALSE; break;
#endif

           case 'X' : 
               if (ch1 == 'T') { gcTimer = optvalue(optarg, "XT"); break; }
               if (ch1 == 'M') { gcMemory = optvalue(optarg, "XM"); break; }
               if (ch1 == 'G') { gcTicksByStep = optvalue(optarg, "XG"); break; }
               if (ch1 == 'A') { gcAppByStep = optvalue(optarg, "XA"); break; }
               if (ch1 == 'V') { gcVerboseLevel = optvalue(optarg, "XV"); break; }

           default:   printf("OFORTH : Unknown option --%c\n", ch); exit(KO);  
         }
        optind++; continue;
       }
    else {
      if (optind >= MAXARGS) return KO;
      argIndexes[optind] = '1';
      nbArguments++;
      }
    if (cp[0] != '-' && optind > 0) { inputFile = cp; }
    optind++;
    }
   return OK;
}
//*******************************************************************
// Main function
//*******************************************************************

#ifdef __OWIN__
__declspec( dllexport )
int WINAPI WinMain(HINSTANCE h, HINSTANCE p, LPSTR cm, int show)
{
int argc = _argc;
char **argv = _argv;
#endif

#ifdef __OLNX__
int main(int argc, char **argv, char **env)
{
int h = 0; 
int show = 0;
#endif

char *s, *path, *path2;
char ofpath[256];
int i, ret;
CThread *thread;
vint size, nbPaths;

   // Initialize default values
   tickInit = getTick(OK, NULL);
   srand(tickInit);
   stdInput      = (FILE *) FTOOINT(stdin);
   gcTimer       = 120;
   gcMemory      = 1024;
   gcTicksByStep = GC_TICK_GC;
   gcAppByStep   = GC_TICK_MUTATOR;
   gcVerboseLevel = 0;

   if (ret_args(argc, argv) == KO)
      { printf("Max number of arguments reached\n"); exit(KO); }

   if (mainDsSize < DEFAULTDSMAIN) mainDsSize = DEFAULTDSMAIN;

   thread = lastThread = mainThread; 
   thread->thread = NULL;
   thread->persistentPage = NULL;
   thread->persistentFree = 0;
   thread->resumableTasks = thread->resumeLast = NULL; 
   thread->sleepingTasks = NULL;
   thread->channels = NULL;
   thread->heapSize = 0;

   // Create main task (interpreter)
   mainTask->awakeOccured = KO;
   mainTask->dsSize = mainDsSize;
   mainTask->rs = mainTask->rsbase = NULL;

   // Continue main thread
   thread->id = workerId; workerId++;
   thread->state = THREAD_RUNNING;
   thread->gcTicks = YIELD_TICKS;
   thread->gcPhase = GC_NOGC;
   thread->currentTask = mainTask;

   thread->raws    = (Raw **)     allocDictCell(thread, RAWSLOTS);
   thread->all_obj = (RawPage **) allocDictCell(thread, MAXATTRIBUTE + 1); 
   thread->cur_obj = (RawPage **) allocDictCell(thread, MAXATTRIBUTE + 1); 
   thread->gcFreeIndex = (vint *) allocDictCell(thread, MAXATTRIBUTE + 1);

   thread->markStackBase = 
        (MyObject **) allocRawMem(thread, MAX_MARKSTACK * sizeof(Cell));  
   thread->markStack     = thread->markStackBase;
   thread->markStackUp   = thread->markStack + (MAX_MARKSTACK - 1);

   // Inner for main task
   size = (mainTask->dsSize + 10) * sizeof(Cell);
   mainTask->dsup = ((PCell) allocRawMem(thread, size)) + 10;
   mainTask->dsbase = mainTask->dsup + mainTask->dsSize;
   *(mainTask->dsbase - 1) = nil; *(mainTask->dsbase - 2) = nil;
   mainTask->dsbase -= 2; 
   mainTask->ds = mainTask->dsbase;
   mainTask->rsMaxSize = 0;
   mainTask->codePage = NULL;
   mainTask->codeBuffer = NULL;

   // Oforth VM Initialization
   ovm = (OforthVM *) allocRawMem(thread, sizeof(OforthVM));

   ovm->state = VM_RUN;
   ovm->nbWorkers = ovm->nbWorkersToSynchronize = 1;
   ovm->nbWorkersSnapshot = ovm->nbWorkersMarking = 0;
   ovm->currentThreadGC = NULL;

   thread->vm = ovm;
   thread->ds      = mainTask->ds;
   // dbg("thread0 = %x dsmain = %x\n", thread, thread->ds);
   thread->dsbase  = mainTask->dsbase;
   thread->dsupmax = mainTask->dsup;

   // Initialize all built-in classes, functions and methods
   init_image(thread, (vint) h, (vint) show, argc, argv);

   // Update main task
   mainTask->runnable = (MyObject *) loadOforth;
   mainTask->natural  = allocDictInteger(thread, 3);

#ifdef __FLOAT_ENABLE__
   mainTask->f.objptr = (Float->classPtr | MSKPERSISTENT) & MSKMODERUN;
#endif


   // Creates nzero + thread temporary natural
   nzero = (MyNatural *) allocDictInteger(thread, 1);
   nzero->sign = 1;
   nzero->buf[0] = 0;

   mainTry->objptr = (Integer->classPtr | MSKPERSISTENT) & MSKMODERUN;
   mainTry->level = OONE;
   mainTry->whenAddr = (void *) sysThreadException;

   stdoutFile = (MyFile *) allocDictObject(thread, File, sizeof(MyFile));
   stdoutFile->h = stdout;
   stdoutFile->name = allocDictString(thread, "stdout", KO);
   stdoutFile->encoding = FILE_UTF8;
   constant(thread, "System.Out", (Cell) stdoutFile);

   stderrFile = (MyFile *) allocDictObject(thread, File, sizeof(MyFile));
   stderrFile->h = stderr;
   stderrFile->name = allocDictString(thread, "stderr", KO);
   stderrFile->encoding = FILE_UTF8;
   constant(thread, "System.Err", (Cell) stderrFile);

#ifdef __OWIN__
int res;
DWORD read, mode;
INPUT_RECORD rec[1];
LARGE_INTEGER tf;

   STDIN  = GetStdHandle(STD_INPUT_HANDLE);
   STDOUT = GetStdHandle(STD_OUTPUT_HANDLE);

   res = PeekConsoleInput(STDIN, rec, 1, &read);

   if (res != 0) {
      consoleMode = OK;
      GetConsoleMode(STDIN, &mode);
      SetConsoleMode(STDIN, mode & ~(ENABLE_ECHO_INPUT | ENABLE_LINE_INPUT) );
      SetConsoleOutputCP(CP_UTF8);
      }

#ifdef __TCP_ENABLE__
   if (SockInit != SockInitOk) { printf("Sock initialization failed...\n"); exit(KO); }
#endif

#ifdef __MTHREAD_ENABLE__
   nbProcs = atoi(getenv("NUMBER_OF_PROCESSORS"));
#endif

   if (QueryPerformanceFrequency(&tf) != 0) 
      tickFrequency = 1000000.0 / tf.QuadPart;

   constant(thread, "System.WIN?", (Cell) OTRUE);
   constant(thread, "System.LINUX?", (Cell) OFALSE);
   constant(thread, "System.MAC?", (Cell) OFALSE);

#endif

#ifdef __OLNX__
struct termios nowait_ts;
 
   if (isatty(fileno(stdin))) {
      tty = stdin;
      consoleMode = OK;
      tcgetattr(STDIN_FILENO, &original_ts);
      nowait_ts = original_ts;

      nowait_ts.c_lflag &= ~(ECHO | ISIG | ICANON | NOFLSH);

      nowait_ts.c_cc[VMIN] = 0;
      nowait_ts.c_cc[VTIME] = 0;
      tcsetattr(STDIN_FILENO, TCSANOW, &nowait_ts);
      fcntl(STDIN_FILENO, F_GETFL, 0);
      }


#ifdef __MTHREAD_ENABLE__
   nbProcs = sysconf(_SC_NPROCESSORS_ONLN);
   pthread_attr_init(&commonAttr);
   pthread_attr_setdetachstate(&commonAttr, PTHREAD_CREATE_DETACHED);
#endif

   tickFrequency = 1.0;

   constant(thread, "System.WIN?", (Cell) OFALSE);

#if defined(__OLNX32__) || defined (__OLNX64__)
   constant(thread, "System.LINUX?", (Cell) OTRUE);
   constant(thread, "System.MAC?", (Cell) OFALSE);
#endif

#if defined(__OMAC32__) || defined (__OMAC64__)
   constant(thread, "System.LINUX?", (Cell) OFALSE);
   constant(thread, "System.MAC?", (Cell) OTRUE);
#endif

#endif

   // Interpreter & console
   nextWord = allocPersistentString(thread, MAXNAME + 4);
   cmdStdin = allocPersistentString(thread, 1024);

   if (consoleMode == OK && interpretMode == OK) {
      mainConsole = (MyConsole *) allocDictObject(thread, Console, sizeof(MyConsole));
      PUSHDS(mainConsole); sysResourceInit(thread);
      mainConsole->selectMode = OK;
      mainConsole->cmd = cmdStdin; 
      replStr = allocDictString(thread, "repl", KO);
      setvar(thread, _CONSOLE, (Cell) mainConsole);
      }

   srand(getTick(OK, NULL));

   // Create list of Oforth paths
   if ((s = getenv("OFORTH_PATH")) == NULL) { 
      printf("OFORTH_PATH variable is not set. See INSTALL.TXT file.\n"); 
      exit(KO); 
      }

   allPath = (MyArray *) allocDictObject(thread, Array, sizeof(MyArray));
   allPath->buf = (Cell *) allocRawMem(thread, MAXPATH * sizeof(Cell));
   
   nbPaths = 0;
   do {
      while(*s == ' ') s++;
      path = path2 = strchr(s, PATH_FORMAT);
      if (path2 == NULL) path = s + strlen(s);
      size = path - s;
      if (size <= 2 || size > 250) 
         { printf("OFORTH_PATH error : length error\n"); exit(KO); }

      strncpy(ofpath, s, size);
      do { ofpath[size] = '\0'; size--; }
      while(size > 0 && (ofpath[size] == '/' || ofpath[size] == '\\')) ;

      if (size <= 2 || size > 250) 
         { printf("OFORTH_PATH error : length error\n"); exit(KO); }

      // dbg("NewPath : %s\n", ofpath);
      allPath->buf[nbPaths] = (Cell) allocDictString(thread, ofpath, OK);
      nbPaths++;
      s = path + 1; 
      }
   while(path2 != NULL && nbPaths < MAXPATH - 1);

   allPath->free = allPath->buf + nbPaths;
   allPath->objptr &= MSKNOTMUTABLE;

   // Create command line arguments
   allArgs = (MyArray *) allocDictObject(thread, Array, sizeof(MyArray));
   allArgs->buf  = (PCell) allocRawMem(thread, nbArguments * sizeof(Cell));
   allArgs->free = allArgs->buf + nbArguments;

   ret = 0;
   for(i = 0; i < MAXARGS; i++) {
      if (argIndexes[i] == '1') {
         (allArgs->buf)[ret] = (Cell) allocDictString(thread, argv[i], KO);
         ret++;
         }
      }

   allArgs->objptr &= MSKNOTMUTABLE;
    

   // Constants
   constant(thread, "CELLSIZE", (Cell) TOOINT(sizeof(Cell)));

   constant(thread, "System.VERSION",(Cell) allocDictString(thread,OFORTH_VERSION,KO));
   constant(thread, "System.TEST?", (Cell) checkState);
   constant(thread, "System.OPTIMIZE?", (Cell) optimMod);

   constant(thread, "System.ARGS", (Cell) allArgs);
   constant(thread, "System.PATHS", (Cell) allPath);

#ifdef __x8632__
   constant(thread, "Assembler",(Cell)allocDictString(thread,"lang/compiler/asmX8632.of",KO));
#endif

   constant(thread, "System.FLOAT?", 
#ifdef __FLOAT_ENABLE__
       (Cell) OTRUE);
#else 
       (Cell) OFALSE);
#endif

   constant(thread, "System.TCP?", 
#ifdef __TCP_ENABLE__
       (Cell) OTRUE);
#else 
       (Cell) OFALSE);
#endif

   constant(thread, "System.DYNLIB?", 
#ifdef __DYN_ENABLE__
       (Cell) OTRUE);
#else 
       (Cell) OFALSE);
#endif

   constant(thread, "System.MTHREAD?", 
#ifdef __MTHREAD_ENABLE__
       (Cell) OTRUE);
#else 
       (Cell) OFALSE);
#endif

   constant(thread, "System.DEBUG?", 
#ifdef __DBG__
       (Cell) OTRUE);
#else 
       (Cell) OFALSE);
#endif


   // Interpreter variables
   STATE             = variable(thread, "STATE");
   setvar(thread, STATE, (Cell) OZERO);

   allBlocks         = variable(thread, "_NESTEDBLOCKS");

   SOURCE            = variable(thread, "SOURCE");
   SOURCETYPE        = variable(thread, "SOURCE-TYPE");
   SOURCENAME        = variable(thread, "SOURCE-NAME");
   SOURCELINE        = variable(thread, "SOURCE-LINE"); 
   SOURCEINDEX       = variable(thread, "SOURCE-INDEX"); 

   CPBEFORELASTCALL  = variable(thread, "_CPBEFORELASTCALL");
   CPLASTCALL        = variable(thread, "_CPLASTCALL");
   LASTCALL          = variable(thread, "_LASTCALL");

   // Initialize GC events timers
   GCLOCAL  = gcTimer * 1000;
   GCGLOBAL = GCLOCAL * 5;
   nextLocalGC = nextGlobalGC = nextInitTick = 0;
   
   // Input characters behavior
   // If CPREFIX, stop reading, word is this character.
   // If CSUFFIX, word characters stops before reading this character
   for(i = 0; i < 0x80; i++) InputChar[i] = 0;
   InputChar[0]    =           CSUFFIX ;
   InputChar['\n'] =           CSUFFIX ;
   InputChar['\r'] =           CSUFFIX ;
   InputChar['\t'] =           CSUFFIX ;
   InputChar[' ']  =           CSUFFIX ;
   InputChar['{']  = CPREFIX | CSUFFIX ;
   InputChar['}']  = CPREFIX | CSUFFIX ;
   InputChar['[']  = CPREFIX | CSUFFIX ;
   InputChar[']']  = CPREFIX | CSUFFIX ;
   InputChar[',']  = CPREFIX | CSUFFIX ;
   InputChar['(']  = CPREFIX | CSUFFIX ;
   InputChar[')']  = CPREFIX | CSUFFIX ;
   InputChar['|']  = CPREFIX | CSUFFIX ;
   InputChar['#']  = CPREFIX ;
   InputChar['\"'] = CPREFIX ;
   InputChar['\''] = CPREFIX ;
   InputChar['$']  = CPREFIX ;
   InputChar['@']  = CPREFIX ;
   InputChar[';']  = CPREFIX ;
   InputChar['\\'] = CPREFIX ;

   // Create workers
#ifdef __MTHREAD_ENABLE__
   if (maxWorkers != 0) { ovm->nbWorkersMax = maxWorkers; }
   else                 { ovm->nbWorkersMax = nbProcs; }

   for(i = 0; i < ovm->nbWorkersMax - 1; i++) { newWorker(thread); }
#endif

   constant(thread, "System.CORES", (Cell) TOOINT(nbProcs));
   constant(thread, "System.WORKERS", (Cell) TOOINT(ovm->nbWorkersMax));

   // Now, create first command and run it into the main thread
   s = (char *) cmdStdin->buf;

   strcpy(s, "\"oforth.of\" load ");

   if (inputFile != NULL) {
      if (strlen(inputFile) > MAXNAME) 
          { perror("File name is too long.\n"); exit(KO); }

      sprintf( (char *) nextWord->buf, "\"%s\" load ", inputFile);
      strcat(s, (char *) nextWord->buf);
      }

   if (cmd != NULL) {
      if (strlen(cmd) > 500 ) { printf("command line is too long.\n"); exit(KO); }
      strcat(s, cmd);
      }

   thread->resumableTasks = mainTask; 
   thread->currentTask = NULL;
   mainTask->state = TASK_RUN;
   sysThread(thread);
   return OK;
}
//*******************************************************************

