
//*****************************************************************************
//*****************************************************************************
// OFORTH : assembler preprocessor
//*****************************************************************************
//*****************************************************************************

/*
  Copyright (c) 2014, 2017 Franck Bensusan. All rights reserved.

  This file is part of Oforth sources engine.

  Oforth is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation, either version 3
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, see http://www.gnu.org/licenses/.
*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define	OK	0
#define	KO 1

//*******************************************************************
int myAtEnd(FILE *f)
//*********************
{
   if (feof(f)) return OK;
   return KO;
}
//******************************************************************
void strcatc(char *str, char c)
//******************************
{
   while (*str) { str ++; }
   *str = c; *(str+1) = '\0';
}
//*******************************************************************
void evaluate(char *in)
//*********************
{
char out[255];
char *s, *t;
int fin = KO;

   while (fin == KO) {
      fin = OK; 
      if ((s = strstr(in, "-=")) != NULL) {
         *out = '\0';
         *s = '\0'; strcpy(out, "sub("); strcat(out, s+2); strcat(out, ", ");
         strcat(out, in); strcat(out, ")"); strcpy(in, out); fin = KO;
         }

      if ((s = strstr(in, "+=")) != NULL) {
         *out = '\0';
         *s = '\0'; strcpy(out, "add("); strcat(out, s+2); strcat(out, ", ");
         strcat(out, in); strcat(out, ")"); strcpy(in, out); fin = KO;
         }

      if ((s = strstr(in, "&=")) != NULL) {
         *out = '\0';
         *s = '\0'; strcpy(out, "and("); strcat(out, s+2); strcat(out, ", ");
         strcat(out, in); strcat(out, ")"); strcpy(in, out); fin = KO;
         }

      if ((s = strstr(in, "|=")) != NULL) {
         *out = '\0';
         *s = '\0'; strcpy(out, "or("); strcat(out, s+2); strcat(out, ", ");
         strcat(out, in); strcat(out, ")"); strcpy(in, out); fin = KO;
         }

      if ((s = strchr(in, '*')) != NULL && (strstr(in, "call ") == NULL)
         && (strstr(in, "jmp") == NULL)) {
         *out = '\0';
         *s = '\0'; s++; strcat(out, in); 
         if (*s != '(') { 
            strcat(out, "at("); 
            while(*s != ' ' && *s != ')' && *s != ';'&& *s && *s != ',' ) 
               { strcatc(out, *s); s++;}
            strcatc(out, ')'); strcat(out, s);
            }
         else { 
            t = s+1; 
            while(*t != '+' && *t != '-') { t++; } 
            if (*t == '+') { strcat(out, "at"); } else { strcat(out, "bt"); }
            *t = '\0'; t++; 
            strcatc(out, *t); strcatc(out, '(');
            strcat(out, s+1); strcatc(out, ')'); strcat(out, t+2);
            }; 
         strcpy(in, out); fin = KO;
         }

      if ((s = strchr(in, '=')) != NULL) {
         *out = '\0';
         *s = '\0'; strcpy(out, "mov("); strcat(out, s+1); strcat(out, ", ");
         strcat(out, in); strcat(out, ")"); strcpy(in, out); fin = KO;
         }

      if ((s = strstr(in, "->")) != NULL) {
         *out = '\0';
         *s = '\0'; strcpy(out, "stm("); strcat(out, in); strcat(out, ", ");
         strcat(out, s+2); strcat(out, ")"); strcpy(in, out); fin = KO;
         }

      if ((s = strstr(in, "<-")) != NULL) {
         *out = '\0';
         *s = '\0'; strcpy(out, "ldm("); strcat(out, in); strcat(out, ", ");
         strcat(out, s+2); strcat(out, ")"); strcpy(in, out); fin = KO;
         }
      }
}
//******************************************************************
int readNextInstr(FILE *f, char *cmd)
//***********************************
{
char c;
int r = KO; 

   *cmd = '\0';
   while(1) {
      c = fgetc(f); 
      if (c == '!') {
         while(1) {
            c = fgetc(f);
            if (c == EOF || c == '\n') break;
            }
         }
      if (c == EOF || c == '\n' || c == ';') return r;
      r = OK; strcatc(cmd, c);
      }
}
//******************************************************************
int analyse(FILE *f, FILE *g)
//***************************
{
char cmd[255];
int i;

	while (myAtEnd(f) == KO) {
      i = readNextInstr(f, cmd);
      if (i == OK) { evaluate(cmd); fputs(cmd, g); fputc('\n', g); }
		}
	return OK;
}
//******************************************************************
int main(int argc, char **argv)
//*****************************
{
FILE *f, *g;
char cible[20], *pc;


	if(argc != 2) {
	   printf("ERR : Syntax : preproc file.ps file.asm\n"); 
      exit(KO); 
      }

	strcpy(cible, argv[1]);
	pc = cible;
	while (*pc != '.' && *pc != '\0') { pc++; }
	
	if (*pc == '\0' || *(pc+1) != 'p' || *(pc+2) != 's' ) {
		printf("PREPROC : ERR : Bad input file\n"); exit(KO); }

	*(pc+1) = 'S'; *(pc+2) = '\0';
	
	f = fopen(argv[1], "r");
	if (f == (FILE *) NULL) 
      { printf("ERR : Can't open <%s> for input\n", argv[1]); exit(KO); }

	g = fopen(cible, "w");
	if (f == (FILE *) NULL) 
      { printf("ERR : Can't open <%s> for output\n", cible); exit(KO); }
	
	if (analyse(f, g) == KO) 
		{ printf("PREPROC : ERR : Error while analysing\n"); exit(KO); }

	fclose(f); fclose(g);

	exit(OK);
}
